
local addon = LibStub("AceAddon-3.0"):NewAddon("PetLeash", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
_G.PetLeash = addon

local L	= LibStub("AceLocale-3.0"):GetLocale("PetLeash")

-- Binding globals
BINDING_HEADER_PETLEASH = "PetLeash"
BINDING_NAME_PETLEASH_SUMMON = L["Summon Another Pet"]
BINDING_NAME_PETLEASH_DESUMMON = L["Desummon Pet"]
BINDING_NAME_PETLEASH_TOGGLE = L["Toggle Non-Combat Pet"]
BINDING_NAME_PETLEASH_CONFIG = L["Open Configuration"]

-- Default DB
local defaults = {
	profile = {
        enable = true,
        enableInCombat = false,
        disableWhileStealthed = true,
        disableForQuestItems = true,
        waitTimer = 3,
        selectedQuestItems = {}, -- "item:xx" = true, disable.  if false, use check
        autoSwitchTimer = false
    },
    char = {
        ignore_pets = {}       -- [spellid] = true, hide.  if false, don't hide
    }
}

-- config

local function config_toggle_get(info) return addon.db.profile[info[#info]] end
local function config_toggle_set(info,v) addon.db.profile[info[#info]] = v end

local config_autoSwitchTimer_oldval = 30

local options = {
    name = "PetLeash",
    handler = PetLeash,
    type = 'group',
    args = {
        general = {
            name = GENERAL,
            type = 'group',
            args = {
                enable = {
                    type = "toggle",
                    name = ENABLE,
                    desc = L["Enable Auto-Summon"], 
                    order = 10,
                    width = "full",
                    get = function(info) return addon:IsEnabledSummoning() end,
                    set = function(info,v) addon:EnableSummoning(v) end,
                },
                enableInCombat = {
                    type = "toggle",
                    name = L["Enable In Combat"],
                    order = 11,
                    width = "double",
                    get = config_toggle_get,
                    set = config_toggle_set
                },
                disableWhileStealthed = {
                    type = "toggle",
                    name = L["Disable While Stealthed or Invisible"],
                    order = 12,
                    width = "double",
                    get = config_toggle_get,
                    set = config_toggle_set
                },
                disableForQuestItems = {
                    type = "toggle",
                    name = L["Disable For Special Items"],
                    desc = L["Disable when special items that summon pets are detected.  This includes quest items and hats."],
                    order = 13,
                    width = "double",
                    get = config_toggle_get,
                    set = config_toggle_set
                },
                questItemsSelector = {
                    type = "multiselect",
                    name = L["Special Items:"],
                    order = 14,
                    style = "dropdown",
                    disabled = function()
                        return not addon.db.profile.disableForQuestItems
                    end,
                    values = function()
                        local r = {}
            
                        for item,func in pairs(addon.quest_items) do
                            r[item] = addon:GetQuestItemName(item)
                        end
            
                        return r
                    end,
                    get = function(info, key)
                        return not addon.db.profile.selectedQuestItems[key]
                    end,
                    set = function(info, key, value)
                        addon.db.profile.selectedQuestItems[key] = not value 
                    end,
                },
                waitTimerValue = {
                    type = "range",
                    name = L["Wait Time (Seconds)"],
                    desc = L["How long must pass before a player is considered idle enough to summon a pet."],
                    order = 15,
                    min = 1,
                    step = .5,
                    bigStep = 1,
                    max = 30,
                    get = function()
                        return addon.db.profile.waitTimer
                    end,
                    set = function(info,v)
                        addon.db.profile.waitTimer = v
                    end  
                },
                --
                autoSwitchTimerHeader = {
                    type = "header",
                    name = L["Timed Pet Auto Switch"],
                    order = 16,
				},
                autoSwitchTimerEnable = {
                    type = "toggle",
                    name = L["Enable Auto Switch"],
                    order = 17,
                    get = function() return addon.db.profile.autoSwitchTimer end,
                    set = function(info,v)
                        if(v) then
                            if(not addon.db.profile.autoSwitchTimer) then
                                addon.db.profile.autoSwitchTimer = config_autoSwitchTimer_oldval
                            end
                        else
                            config_autoSwitchTimer_oldval = addon.db.profile.autoSwitchTimer
                            addon.db.profile.autoSwitchTimer = v
                            addon:StartAutoSwitchTimer()
                        end
                    end
                },
                autoSwitchTimerValue = {
                    type = "range",
                    name = L["Seconds Between Switch"],
                    order = 18,
                    min = 30,
                    step = 1,
                    bigStep = 60,
                    max = 3600,
                    disabled = function() return not addon.db.profile.autoSwitchTimer end,
                    get = function()
                        return addon.db.profile.autoSwitchTimer or config_autoSwitchTimer_oldval
                    end,
                    set = function(info,v)
                        addon.db.profile.autoSwitchTimer = v
                        addon:StartAutoSwitchTimer()
                    end
                },
            }
        },
        pets = {
            type = "group",
            name = L["Enabled Pets"],
            order = 10,
            cmdHidden = true,
            args = {}
        },
    },
}

local options_slashcmd = {
    name = "PetLeash Slash Command",
    handler = PetLeash,
	type = "group",
	order = -2,
	args = {
        config = {
            type = "execute",
            name = L["Open Configuration"],
            dialogHidden = true,
            order = 1,
            func = function(info) addon:OpenOptions() end
        },
        resummon = {
            type = "execute",
            name = L["Summon Another Pet"],
            desc = L["Desummon your current pet and summon another pet.  Enable summoning if needed."],
            order = 20,
            func = function(info) addon:ResummonPet() end
        },
        desummon = {
            type = "execute",
            name = L["Desummon Pet"],
            desc = L["Desummon your currently summoned pet.  Disable summoning."],
            order = 21,
            func = function(info) addon:DesummonPet() end
        },
        togglePet = {
            type = "execute",
            name = L["Toggle Non-Combat Pet"],
            order = 22,
            func = function(info) addon:TogglePet() end
        },
        enable = options.args.general.args.enable,
        enableInCombat = options.args.general.args.enableInCombat,
        disableWhileStealthed = options.args.general.args.disableWhileStealthed,
        disableForQuestItems = options.args.general.args.disableForQuestItems,
	},
}

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

function addon:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("PetLeashDB", defaults)

    self.usable_pets = {}   -- spellid, spellid
    self.pet_map = {}       -- spellid -> {id,name} (complete)
    self.player_invisible = false
    
    AceConfig:RegisterOptionsTable(self.name, options)
    self.optionsFrame = AceConfigDialog:AddToBlizOptions(self.name, self.name, nil, "general")
    self.optionsFrame.Pets = AceConfigDialog:AddToBlizOptions(self.name, L["Enabled Pets"], self.name, "pets")
    self.optionsFrame.About = LibStub("LibAboutPanel").new(self.name, self.name)
    AceConfig:RegisterOptionsTable(self.name .. "SlashCmd", options_slashcmd, {"petleash", "pl"})
   
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
    self:RegisterEvent("COMPANION_UPDATE")
    self:RegisterEvent("COMPANION_LEARNED")
    self:RegisterEvent("UPDATE_STEALTH")
    self:RegisterEvent("UNIT_AURA")

    self:LoadPets()                             -- attempt to load pets (might fail)
    
    self:InitBroker()
end

function addon:IsEnabledSummoning()
    return self.db.profile.enable
end

function addon:EnableSummoning(v)
    local oldv = self.db.profile.enable

    if((not oldv) ~= (not v)) then
        self.db.profile.enable = v
        
        -- TODO: is there a better way to trigger config update?
        AceConfigRegistry:NotifyChange("PetLeash")
    end
end

function addon:OpenOptions()
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
end

-- utility functions

local function HasCompanion(pettype)
    for id = 1,GetNumCompanions(pettype) do
        local _,_,_,_,issum = GetCompanionInfo(pettype, id)
        if(issum) then
            return id
        end
    end
    return nil
end

local function IsMounted()
    return _G.IsMounted()
    --return HasCompanion("MOUNT")
end

local function IsPetted(id)
    if(id) then
        local _,_,_,_,issum = GetCompanionInfo("CRITTER", id)
        if(issum) then
            return id
        end
    else
        return HasCompanion("CRITTER")
    end
end
addon.IsPetted = IsPetted           -- expose (mostly for debugging)

local function IsCasting()
    return (UnitCastingInfo("player") or UnitChannelInfo("player"))
end

local function IsDrinkingOrEating()
    local drink_name = GetSpellInfo(430)
    local eat_name = GetSpellInfo(433)
    
    if(UnitAura("player", drink_name, nil, "HELPFUL")) then
        return true
    elseif(UnitAura("player", eat_name, nil, "HELPFUL")) then
        return true
    end
end

local InCombat = InCombatLockdown       -- shorthand

local function CanSummonPet()
    return
        -- are we busy?
            not IsCasting()
        and not IsMounted()
        and not UnitInVehicle("player")
        and not UnitIsGhost("player")
        and not UnitIsDead("player")
        and not UnitOnTaxi("player")
        and not UnitIsFeignDeath("player")
        and not IsFlying()
        and not IsFalling()
        and not IsDrinkingOrEating()
        and HasFullControl()
        -- verify we have pets
        and GetNumCompanions("CRITTER") > 0
        -- gcd check
        and GetCompanionCooldown("CRITTER", 1) == 0
end
addon.CanSummonPet = CanSummonPet       -- expose (mostly for debugging)

-- pet list handling

function addon:LoadPets(updateconfig)
    wipe(self.usable_pets)
    
    for i = 1,GetNumCompanions("CRITTER") do
        local _, name, spellid = GetCompanionInfo("CRITTER", i)
        
        if(not name) then
            return      -- pets not loaded yet?
        end
        
        if(not self.db.char.ignore_pets[spellid]) then
            table.insert(self.usable_pets, spellid)
        end

        if(not self.pet_map[spellid]) then
            self.pet_map[spellid] = {}
        end
        self.pet_map[spellid].id = i
        self.pet_map[spellid].name = name 
    end
    
    if(updateconfig == nil or updateconfig) then
        self:UpdateConfigTables()
    end
    
    self:UpdatePetTimer()
end

function addon:UpdateConfigTables()
    local args = options.args.pets.args
    
    wipe(args)
    
    for i = 1,GetNumCompanions("CRITTER") do
        local _, name, spellid = GetCompanionInfo("CRITTER", i)
        
        args[tostring(spellid)] = {
            type = "toggle",
            name = name,
            order = 1,
            get = "Config_PetToggle_Get",
            set = "Config_PetToggle_Set"
        }
    end
    
    -- Config Tables changed!
    AceConfigRegistry:NotifyChange("PetLeash")
end

function addon:Config_PetToggle_Set(info, v)
    if(v) then
        self.db.char.ignore_pets[tonumber(info[#info])] = nil
    else
        self.db.char.ignore_pets[tonumber(info[#info])] = true
    end
    
    self:LoadPets(false)
end

function addon:Config_PetToggle_Get(info)
    return not self.db.char.ignore_pets[tonumber(info[#info])]
end

-- events

function addon:PLAYER_ENTERING_WORLD()
    -- reload hijinks: maybe we have a pet out already!
    if(#self.usable_pets > 0 and IsPetted()) then
        self:StartAutoSwitchTimer()
    end
end

function addon:COMPANION_UPDATE(event, ctype)
    if(ctype == nil) then
        self:LoadPets()
    elseif(ctype == "CRITTER") then
        -- TODO: pet was shown or hidden
        self:UpdatePetTimer()
    end
end

function addon:COMPANION_LEARNED()
    self:LoadPets()
end

function addon:UPDATE_STEALTH()
    if(IsStealthed()) then
        if(self.db.profile.disableWhileStealthed) then
            -- desummon pet, but don't disable completely
            self:DesummonPet(true)
        end
    else
        self:UpdatePetTimer()
    end
end

function addon:UNIT_AURA(event, unit)
    if(unit ~= "player") then
        return
    end

    -- check for invisibility
    local invis_name, _, invis_texture = GetSpellInfo(66)
    local c_name, _, c_texture = UnitAura("player", invis_name, nil, "HELPFUL")
    if(c_name == invis_name and invis_texture == c_texture) then
        if(not self.player_invisible) then
            self.player_invisible = true
            if(IsPetted() and self.db.profile.disableWhileStealthed) then
                self:DesummonPet(true)
            end
        end
    else
        if(self.player_invisible) then
            self.player_invisible = false
            self:UpdatePetTimer()
        end
    end
end

function addon:UpdatePetTimer()
    local haspet = self:HasPet(true)
    if(not haspet) then
        self:StartPetTimer()
    else
        self:StopPetTimer()
    end
    
end

local countdown
function addon:StartPetTimer()
    if(self.pet_timer) then
        -- stop timer
        self:CancelTimer(self.pet_timer)
    end
    self.pet_timer = self:ScheduleRepeatingTimer("PeriodicCheckPet", 0.5)
    countdown = self.db.profile.waitTimer*2    
end

function addon:StopPetTimer()
    self:CancelTimer(self.pet_timer)
    self.pet_timer = nil
end

function addon:PeriodicCheckPet()
    countdown = countdown - 1
    if(not self:IsPetSummonReady()) then
        -- reset timer
        countdown = self.db.profile.waitTimer*2
    elseif(countdown == 0) then
        self:SummonPet()
        self:StopPetTimer()
    end
end

function addon:IsPetSummonReady()
    if(not self.db.profile.enable) then
        return
    elseif(not self.db.profile.enableInCombat and InCombat()) then
        return
    elseif(self.db.profile.disableWhileStealthed and (IsStealthed() or self.player_invisible)) then
        return
    elseif(self.db.profile.disableForQuestItems and self:HasQuestItem()) then
        return
    end 

    if(CanSummonPet()) then
        return true
    end
end

function addon:HasPet(nocache)
    if(nocache) then
        local pet_id = IsPetted(self.pet_id) or IsPetted()
        self.pet_id = pet_id
        return pet_id
     end
     return self.pet_id
end

function addon:StartAutoSwitchTimer(override_time)
    if(self.switch_timer_handle) then
        self:CancelTimer(self.switch_timer_handle)
        self.switch_timer_handle = nil
    end

    local timer = override_time or self.db.profile.autoSwitchTimer
    if(timer) then
        self.switch_timer_handle = self:ScheduleTimer("AutoSwitchTimer", timer)
    end
end

function addon:AutoSwitchTimer()
    self.switch_timer_handle = nil     -- timer is finished!
    
    self:StartPetTimer()
end

function addon:SummonPet()
    local n = #self.usable_pets
    if(n > 0) then
        local random_spellid = self.usable_pets[math.random(n)]
        CallCompanion("CRITTER", self.pet_map[random_spellid].id)
        
        self:StartAutoSwitchTimer()
    end
end

function addon:DesummonPet(disable)
    DismissCompanion("CRITTER")
    if(not disable or disable == nil) then
        addon:EnableSummoning(false)
    end
end

function addon:ResummonPet()
    self:EnableSummoning(true)

    self:DesummonPet(true)
    self:SummonPet()
end

function addon:TogglePet()
    if(IsPetted()) then
        self:DesummonPet()
    else
        self:EnableSummoning(true)
        if(CanSummonPet()) then
            self:SummonPet()
        end
    end
end
